
#include <SMPFunction.h>
#include <SMPInstr.h>
#include <interfaces/abstract/STARSFunction.h>
#include <interfaces/irdb/STARSFunction.h>
#include <interfaces/irdb/STARSInstruction.h>
#include <interfaces/irdb/STARSInterface.h>
#include <libIRDB-core.hpp>
#include <libIRDB-cfg.hpp>

#include <list>

using namespace std;
using namespace libIRDB;

void STARS_IRDB_Function_t::IRDB_IR_Build_helper
		(
			SMPFunction *func,
			SMPInstr *CurrInst 
		)
{
#if 1
				SMPitype DataFlowType = CurrInst->GetDataFlowType();
				if ((DataFlowType == INDIR_CALL) || (DataFlowType == CALL)) {
					// See if IDA has determined the target of the call.
#if 0                           // Phase ordering: Do this later.
					CurrInst->AnalyzeCallInst(func->GetFirstFuncAddr());
#endif
					STARS_ea_t TargetAddr = CurrInst->GetCallTarget();
					bool LinkedToTarget = (STARS_BADADDR != TargetAddr);
					if (LinkedToTarget) {
						if (0 == TargetAddr) {
							SMP_msg("WARNING: Ignoring NULL call target (unreachable) at %lx\n",
								(unsigned long) CurrInst->GetAddr());
						}
						else {
							if (INDIR_CALL == DataFlowType) {
								pair<set<STARS_ea_t>::iterator, bool> InsertResult = func->IndirectCallTargets.insert(TargetAddr);
								if (InsertResult.second) {
									func->AllCallTargets.push_back(TargetAddr);
								}
							}
							else {
								(void) func->AddDirectCallTarget(TargetAddr);
							}
						}
					}
					if (DataFlowType == INDIR_CALL) {
						func->IndirectCalls = true;
						func->UnresolvedIndirectCalls = (!LinkedToTarget);
					}
				} // end if INDIR_CALL or CALL
				else if (DataFlowType == INDIR_JUMP) {
					func->IndirectJumps = true;
// was if STARS_AUDIT_INDIR_JUMP_XREFS
#if 0
					PreviousIndirJumpAddr = addr;
#endif
				}
				else if (DataFlowType == RETURN) {
					func->HasReturnInst = true;
				}
/*
				else if (CurrInst->IsBranchToFarChunk()) {
					probably doesn't apply to IRDB.
				}	
*/



#endif
}

void STARS_IRDB_Function_t::BuildFuncIR(SMPFunction *func)
{

	cout<<"Building IR for func "<<the_func->GetName()<<endl;

typedef list<SMPInstr*> SMPInstrList_t;

	SMPInstrList_t::iterator FirstInBlock = func->Instrs.end();
	SMPInstrList_t::iterator LastInBlock = func->Instrs.end();
	bool GoodRTL=false;
	func->BuiltRTLs = true;


	// build cfg 
	ControlFlowGraph_t cfg(the_func);

	// sort the blocks into any order where the entry block is first.
	vector<BasicBlock_t*> sorted_blocks;

	sorted_blocks.push_back(cfg.GetEntry());
	for(BasicBlockSet_t::iterator it=cfg.GetBlocks().begin();
		it!=cfg.GetBlocks().end(); ++it)
	{
		BasicBlock_t* block=*it;
		if(cfg.GetEntry() ==  block)	// we did the entry block first.
			continue;
		sorted_blocks.push_back(block);
	}




	bool first_in_func=true;
	for(vector<BasicBlock_t*>::iterator block_it=sorted_blocks.begin();
			block_it!=sorted_blocks.end(); ++block_it)
	{
		list<SMPInstr*> block_id_map;
		block_id_map.clear();
		BasicBlock_t* block=*block_it;
		for(InstructionVector_t::iterator insn_it=block->GetInstructions().begin();
			insn_it!=block->GetInstructions().end(); ++insn_it)
		{
			Instruction_t* irdb_insn=*insn_it;
//			cout<<"Building IR for instruction "<<dec<<irdb_insn->GetBaseID()<<":"<<irdb_insn->getDisassembly()<<endl;
			STARS_IRDB_Instruction_t* interface_insn=new STARS_IRDB_Instruction_t(irdb_insn);
			STARS_InstructionID_t instruction_id=interface_insn->GetID();
			// create instruction
			SMPInstr *CurrInst = new SMPInstr(interface_insn->GetID().GetIDWithinFile());
			CurrInst->Analyze();


			if(first_in_func)
			{
				first_in_func=false;

				// create SSA marker instruction
				SMPInstr *MarkerInst = new SMPInstr(STARS_SSA_MARKER_PSEUDO_ID);
				block_id_map.push_back(MarkerInst);

				MarkerInst->AnalyzeMarker();
				assert(STARS_NN_fnop == MarkerInst->GetIDAOpcode());
				GoodRTL = MarkerInst->HasGoodRTL();
				func->BuiltRTLs = (func->BuiltRTLs && GoodRTL);
				if (!GoodRTL) {
					cout << "Could not build RTLs!" << endl;
				}
				assert(FirstInBlock == func->Instrs.end());
				func->Instrs.push_back(MarkerInst);


				// find all callers.
				//for each instruction id, FromAddr, that might jump 
				//	 (but not fallthrough):
				STARS_IRDB_Interface_t *sGi=dynamic_cast<STARS_IRDB_Interface_t*>
					(global_stars_interface);
				assert(sGi);
				Instruction_t* entry=the_func->GetEntryPoint();
				for(InstructionSet_t::iterator tmp_it=sGi->instruction_preds[entry].begin();
					tmp_it!=sGi->instruction_preds[entry].end(); ++tmp_it)
				{
					Instruction_t* tmp=*tmp_it;
					if(tmp->GetFunction()!=the_func && tmp->GetFallthrough()!=entry)
						// tell stars about the jump
						func->AddCallSource(tmp->GetBaseID());
				}

				if(the_func->GetEntryPoint()->GetIndirectBranchTargetAddress())
					func->PossibleIndirectCallTarget = true;


			}
			block_id_map.push_back(CurrInst);
	
			// do a bunch of stuff that is based on the currently built IR.  why 
			// this needs to be specific to IRDB I don't know.  Ask clc.
			IRDB_IR_Build_helper(func,CurrInst);

			func->BuiltRTLs = (func->BuiltRTLs && CurrInst->HasGoodRTL());
//			CurrInst->Dump();

			// Insert instruction at end of list.
			func->Instrs.push_back(CurrInst);
			
		}

		// find start/end of block

		// get first and last insn in the block.
		FirstInBlock=find( func->Instrs.begin(), func->Instrs.end(), *(block_id_map.begin()));
		LastInBlock =find( func->Instrs.begin(), func->Instrs.end(), *(--block_id_map.end()));
		SMPBasicBlock *NewBlock = new SMPBasicBlock(func, FirstInBlock, LastInBlock);
		func->Blocks.push_back(NewBlock);
		func->BlockCount += 1;


	}

}




bool STARS_IRDB_Function_t::IsInstIDInFunc(STARS_ea_t InstID)
{
	STARS_InstructionID_t id(InstID);
	const STARS_Instruction_t *stars_insn=id.GetInstruction();
	const STARS_IRDB_Instruction_t* interface_insn=dynamic_cast<const STARS_IRDB_Instruction_t*>(stars_insn);
	assert(interface_insn);
	libIRDB::Instruction_t* irdb_insn=(libIRDB::Instruction_t*)interface_insn->GetIRDBInstruction();
	InstructionSet_t::iterator it=the_func->GetInstructions().find(irdb_insn);
	if(it==the_func->GetInstructions().end())
		return false;
	return true;
}


// assert this represents the same function as CurrFunc
// IsIndirectCallTarget if some instruction (might) indirectly
//      branch to this function's entry.
// IsTailCallTarget if some function has a jump to this function's entry
// use IRDB-util::preds class to check that all predecessors are
// call instructions.  How to deal with push/jmp pairs?  ignore for now?
bool STARS_IRDB_Function_t::AnalyzeInstAsCallTarget(SMPFunction *CurrFunc,
		bool &IsIndirectCallTarget,
		bool &IsTailCallTarget)
{
	// Init.
	IsIndirectCallTarget=false;
	IsTailCallTarget=false;

	// Sanity
	assert(the_func);

	// Get Entry.
	Instruction_t* entry=the_func->GetEntryPoint();
	assert(entry);

	// Check if entry is an IBT 
	if(entry->GetIndirectBranchTargetAddress())
		// report appropriately.
		IsIndirectCallTarget=true;

	extern STARS_Interface_t* global_stars_interface;
	// safe convert.
	STARS_IRDB_Interface_t *sGi=dynamic_cast<STARS_IRDB_Interface_t*>( global_stars_interface);
	assert(sGi);


	for(InstructionSet_t::iterator it=sGi->instruction_preds[entry].begin();
		it!=sGi->instruction_preds[entry].end(); ++it)
	{
		Instruction_t* pred=*it;

		// skip any fallthroughs to the entry (weird, but could happen)
		if(pred->GetFallthrough()==entry)
			continue;

		// skip any call instructions that target the branch
		DISASM d;
		pred->Disassemble(d);
		if(d.Instruction.BranchType==CallType)
			continue;

		IsTailCallTarget=true;
	}

	return true;

}

